package de.fhmuenster.masterthesis.serialization;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.xtext.serializer.impl.Serializer;

import com.google.inject.Guice;
import com.google.inject.Injector;

import de.fhmuenster.masterthesis.TestgeneratorDSLRuntimeModule;
import de.fhmuenster.masterthesis.TestgeneratorDSLStandaloneSetup;
import de.fhmuenster.masterthesis.testgeneratorDSL.BPMNDiagram;
import de.fhmuenster.masterthesis.testgeneratorDSL.BooleanVariable;
import de.fhmuenster.masterthesis.testgeneratorDSL.Declaration;
import de.fhmuenster.masterthesis.testgeneratorDSL.DelegateMock;
import de.fhmuenster.masterthesis.testgeneratorDSL.DelegateServiceTaskFlowElement;
import de.fhmuenster.masterthesis.testgeneratorDSL.ExternalServiceMock;
import de.fhmuenster.masterthesis.testgeneratorDSL.ExternalServiceTaskFlowElement;
import de.fhmuenster.masterthesis.testgeneratorDSL.ExternalTopic;
import de.fhmuenster.masterthesis.testgeneratorDSL.Flag;
import de.fhmuenster.masterthesis.testgeneratorDSL.Flow;
import de.fhmuenster.masterthesis.testgeneratorDSL.FlowElement;
import de.fhmuenster.masterthesis.testgeneratorDSL.FlowElementReference;
import de.fhmuenster.masterthesis.testgeneratorDSL.IntVariable;
import de.fhmuenster.masterthesis.testgeneratorDSL.Loop;
import de.fhmuenster.masterthesis.testgeneratorDSL.Mock;
import de.fhmuenster.masterthesis.testgeneratorDSL.MockReference;
import de.fhmuenster.masterthesis.testgeneratorDSL.Model;
import de.fhmuenster.masterthesis.testgeneratorDSL.ServiceTaskFlowElement;
import de.fhmuenster.masterthesis.testgeneratorDSL.StartFlowElement;
import de.fhmuenster.masterthesis.testgeneratorDSL.StringVariable;
import de.fhmuenster.masterthesis.testgeneratorDSL.Test;
import de.fhmuenster.masterthesis.testgeneratorDSL.TestgeneratorDSLFactory;
import de.fhmuenster.masterthesis.testgeneratorDSL.UserTaskFlowElement;
import de.fhmuenster.masterthesis.testgeneratorDSL.Variable;
import de.fhmuenster.masterthesis.testgeneratorDSL.VariableCreationType;
import de.fhmuenster.masterthesis.utils.TestgeneratorDSLUtils;

@SuppressWarnings("restriction")
public class TestgeneratorDSLSerializer {

	private File dslFile;
	private Model model;
	private Serializer serializer;

	public TestgeneratorDSLSerializer(String path) throws IOException {
		dslFile = new File(path);

		// Deserialization
		new TestgeneratorDSLStandaloneSetup().createInjectorAndDoEMFRegistration();
		createOrGetModel(path);

		// Serialization
		Injector injector = Guice.createInjector(new TestgeneratorDSLRuntimeModule());
		serializer = injector.getInstance(Serializer.class);

	}

	private void createOrGetModel(String path) throws IOException {
		ResourceSet rs = new ResourceSetImpl();
		URI uri = URI.createFileURI(path);

		if (!dslFile.exists()) {
			String dslFileDirectory = dslFile.toPath().getParent().toString();
			File dslFileDir = new File(dslFileDirectory);
			dslFileDir.mkdirs();
			dslFile.createNewFile();
		}
		
		Resource resource = rs.getResource(uri, true);
		if(resource.getContents().isEmpty()) {
			model = TestgeneratorDSLObjectCreator.createModel();
			resource.getContents().add(model);
		} else {
			model = (Model) resource.getContents().get(0);
		}
	}

	public BPMNDiagram getBPMNDiagram(String id) {
		Optional<BPMNDiagram> diagramOpt = model.getDeclarations().stream() //
				.filter(decl -> decl instanceof BPMNDiagram) //
				.map(decl -> ((BPMNDiagram) decl)) //
				.filter(diagram -> diagram.getName().equals(id)) //
				.findFirst(); //

		return diagramOpt.isPresent() ? diagramOpt.get() : null;
	}

	public ExternalServiceTaskFlowElement getExternalServiceTaskFlowElement(String name) {
		Optional<ExternalServiceTaskFlowElement> elementOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof ExternalServiceTaskFlowElement) //
				.map(f -> ((ExternalServiceTaskFlowElement) f)) //
				.filter(element -> element.getName().equals(name)) //
				.findFirst(); //
		
		return elementOpt.isPresent() ? elementOpt.get() : null;
	}
	
	public List<ExternalServiceTaskFlowElement> getExternalServiceTaskFlowElements() {
		List<ExternalServiceTaskFlowElement> externalServiceTaskElements = model.getDeclarations().stream() //
				.filter(d -> d instanceof ExternalServiceTaskFlowElement) //
				.map(f -> ((ExternalServiceTaskFlowElement) f)) //
				.collect(Collectors.toList()); //
		
		return externalServiceTaskElements;
	}
	
	public DelegateServiceTaskFlowElement getDelegateServiceTaskFlowElement(String name) {
		Optional<DelegateServiceTaskFlowElement> elementOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof DelegateServiceTaskFlowElement) //
				.map(f -> ((DelegateServiceTaskFlowElement) f)) //
				.filter(element -> element.getName().equals(name)) //
				.findFirst(); //
		
		return elementOpt.isPresent() ? elementOpt.get() : null;
	}
	
	public List<ServiceTaskFlowElement> getDelegateServiceTaskFlowElements() {
		List<ServiceTaskFlowElement> delegateServiceTaskElements = model.getDeclarations().stream() //
				.filter(d -> d instanceof DelegateServiceTaskFlowElement) //
				.map(f -> ((ServiceTaskFlowElement) f)) //
				.collect(Collectors.toList()); //
		
		return delegateServiceTaskElements;
	}
	
	public List<UserTaskFlowElement> getUserTaskFlowElements() {
		List<UserTaskFlowElement> userTaskElements = model.getDeclarations().stream() //
				.filter(d -> d instanceof UserTaskFlowElement) //
				.map(f -> ((UserTaskFlowElement) f)) //
				.collect(Collectors.toList()); //
		
		return userTaskElements;
	}
	
	public List<StartFlowElement> getStartFlowElements() {
		List<StartFlowElement> startFlowElements = model.getDeclarations().stream() //
				.filter(d -> d instanceof StartFlowElement) //
				.map(f -> ((StartFlowElement) f)) //
				.collect(Collectors.toList()); //
		
		return startFlowElements;
	}
	
	public List<FlowElement> getFlowElements() {
		return model.getDeclarations().stream() //
				.filter(decl -> decl instanceof FlowElement) //
				.map(f -> ((FlowElement) f)) //
				.collect(Collectors.toList()); //
	}

	public List<FlowElement> getFlowElements(List<String> ids) {
		List<FlowElement> flowElements = model.getDeclarations().stream() //
				.filter(decl -> decl instanceof FlowElement) //
				.map(f -> ((FlowElement) f)) //
				.filter(f -> ids.contains(f.getName())) //
				.collect(Collectors.toList()); //

		flowElements.sort(new ReferenceComparator(ids));

		return flowElements;
	}

	public List<Flow> getFlows() {
		return model.getDeclarations().stream() //
				.filter(d -> d instanceof Flow) //
				.map(f -> ((Flow) f)) //
				.collect(Collectors.toList()); //
	}
	
	public Flow getFlow(String name) {
		Optional<Flow> flowOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Flow) //
				.map(f -> ((Flow) f)) //
				.filter(flow -> flow.getName().equals(name)) //
				.findFirst(); //
		
		return flowOpt.isPresent() ? flowOpt.get() : null;
	}
	
	public List<Variable> getVariables() {
		return model.getDeclarations().stream() //
			.filter(d -> d instanceof Variable) //
			.map(varDecl -> ((Variable) varDecl)) //
			.collect(Collectors.toList()); //
	}
	
	public List<VariableProposal<?>> getVariableProposals() {
		List<VariableProposal<?>> variableProposals = new ArrayList<>();
		
		List<VariableProposal<String>> stringProposals = model.getDeclarations().stream() //
			.filter(d -> d instanceof StringVariable) //
			.map(varDecl -> ((StringVariable) varDecl)) //
			.map(stringVar -> new VariableProposal<String>(stringVar.getName(), stringVar.getProposals(), String.class))
			.collect(Collectors.toList());
		
		List<VariableProposal<Integer>> intProposals = model.getDeclarations().stream() //
				.filter(d -> d instanceof IntVariable) //
				.map(varDecl -> ((IntVariable) varDecl)) //
				.map(intVar -> new VariableProposal<Integer>(intVar.getName(), intVar.getProposals(), Integer.class))
				.collect(Collectors.toList());
		
		List<VariableProposal<Boolean>> booleanProposals = model.getDeclarations().stream() //
				.filter(d -> d instanceof BooleanVariable) //
				.map(varDecl -> ((BooleanVariable) varDecl)) //
				.map(boolVar -> new VariableProposal<Boolean>(boolVar.getName(), TestgeneratorDSLUtils.mapDSLBoolean(boolVar.getProposals()), Boolean.class))
				.collect(Collectors.toList());
		
		variableProposals.addAll(stringProposals);
		variableProposals.addAll(intProposals);
		variableProposals.addAll(booleanProposals);
		return variableProposals;
	}
	
	public List<Variable> getManuallyCreatedVariables() {
		return model.getDeclarations().stream() //
				.filter(d -> d instanceof Variable) //
				.map(varDecl -> ((Variable) varDecl)) //
				.filter(varDecl -> varDecl.getCreationType().equals(VariableCreationType.MANUALLY))
				.collect(Collectors.toList()); //
	}

	public Mock getMock(String name) {
		Optional<Mock> mockOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Mock) //
				.map(mock -> ((Mock) mock)) //
				.filter(mock -> mock.getName().equals(name)) //
				.findFirst(); //
		
		return mockOpt.isPresent() ? mockOpt.get() : null;
	}
	
	public List<ExternalServiceMock> getExternalServiceMocks(ExternalServiceTaskFlowElement serviceTask) {
		List<ExternalServiceMock> externalServiceMocks = model.getDeclarations().stream() //
				.filter(d -> d instanceof ExternalServiceMock) //
				.map(mock -> ((ExternalServiceMock) mock)) //
				.filter(mock -> mock.getExternalTopic().equals(serviceTask.getExternalTopic())) //
				.collect(Collectors.toList()); //
		
		return externalServiceMocks;
	}
	
	public List<ExternalServiceMock> getExternalServiceMocks() {
		List<ExternalServiceMock> externalServiceMocks = model.getDeclarations().stream() //
				.filter(d -> d instanceof ExternalServiceMock) //
				.map(mock -> ((ExternalServiceMock) mock)) //
				.collect(Collectors.toList()); //
		
		return externalServiceMocks;
	}
	
	public ExternalServiceMock getExternalServiceMock(String name) {
		Optional<ExternalServiceMock> externalServiceMockOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof ExternalServiceMock) //
				.map(mock -> ((ExternalServiceMock) mock)) //
				.filter(mock -> mock.getName().equals(name)) //
				.findFirst(); //
		
		return externalServiceMockOpt.isPresent() ? externalServiceMockOpt.get() : null;
	}
	
	public List<DelegateMock> getDelegateMocks(DelegateServiceTaskFlowElement serviceTask) {
		List<DelegateMock> delegateMocks = model.getDeclarations().stream() //
				.filter(d -> d instanceof DelegateMock) //
				.map(mock -> ((DelegateMock) mock)) //
				.filter(mock -> mock.getServiceTaskReference().getRef().getName().equals(serviceTask.getName())) //
				.collect(Collectors.toList()); //
		
		return delegateMocks;
	}
	
	public List<DelegateMock> getDelegateMocks() {
		List<DelegateMock> delegateMocks = model.getDeclarations().stream() //
				.filter(d -> d instanceof DelegateMock) //
				.map(mock -> ((DelegateMock) mock)) //
				.collect(Collectors.toList()); //
		
		return delegateMocks;
	}
	
	public DelegateMock getDelegateMock(String name) {
		Optional<DelegateMock> delegateMockOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof DelegateMock) //
				.map(mock -> ((DelegateMock) mock)) //
				.filter(mock -> mock.getName().equals(name)) //
				.findFirst(); //
		
		return delegateMockOpt.isPresent() ? delegateMockOpt.get() : null;
	}
	
	public void deleteMock(String name) {
		Optional<Mock> mockOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Mock) //
				.map(mock -> ((Mock) mock)) //
				.filter(mock -> mock.getName().equals(name)) //
				.findFirst(); //
		
		if(mockOpt.isPresent()) {
			model.getDeclarations().remove(mockOpt.get());
		}
	}
	
	public boolean hasMockReferences(String mockRef) {
		return model.getDeclarations().stream() //
				.filter(d -> d instanceof Test) //
				.map(test -> ((Test) test)) //
				.filter(test -> containsMockReference(test.getMocks(), mockRef))
				.findAny() //
				.isPresent(); //
	}
	
	public List<Test> getAndDeleteOldMockReferences(String oldMockRef) {
		List<Test> mockRefTests = model.getDeclarations().stream() //
			.filter(d -> d instanceof Test) //
			.map(test -> ((Test) test)) //
			.filter(test -> containsMockReference(test.getMocks(), oldMockRef))
			.collect(Collectors.toList()); //
		
		mockRefTests.stream() //
			.map(test -> test.getMocks()) //
			.forEach(references -> references.removeIf(ref -> ref.getRef().getName().equals(oldMockRef))); //
		
		return mockRefTests;
	}
	
	private boolean containsMockReference(EList<MockReference> mockReferences, String oldMockRef) {
		return mockReferences.stream() //
			.filter(ref -> ref.getRef().getName().equals(oldMockRef)) //
			.findAny() //
			.isPresent(); //
	}
	
	public void addMockToTest(Test test, Mock mock) {
		MockReference mockReference = TestgeneratorDSLFactory.eINSTANCE.createMockReference();
		mockReference.setRef(mock);
		
		test.getMocks().add(mockReference);
	}
	
	public Test getTest(String name) {
		Optional<Test> testOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Test) //
				.map(test -> ((Test) test)) //
				.filter(test -> test.getName().equals(name)) //
				.findFirst(); //
		
		return testOpt.isPresent() ? testOpt.get() : null;
	}
	
	public void deleteTest(String name) {
		Optional<Test> testOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Test) //
				.map(test -> ((Test) test)) //
				.filter(test -> test.getName().equals(name)) //
				.findFirst(); //
		
		if(testOpt.isPresent()) {
			model.getDeclarations().remove(testOpt.get());
		}
	}
	
	public void deleteFlow(String name) {
		Optional<Flow> flowOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Flow) //
				.map(flow -> ((Flow) flow)) //
				.filter(flow -> flow.getName().equals(name)) //
				.findFirst(); //
		
		if(flowOpt.isPresent()) {
			model.getDeclarations().remove(flowOpt.get());
		}
	}
	
	public Map<String, Long> getFlowTestCount() {
		return model.getDeclarations().stream()
			.filter(d -> d instanceof Test) //
			.map(d -> ((Test) d).getFlowReference().getRef().getName()) //
			.collect(Collectors.groupingBy(Function.identity(), Collectors.counting()));
	}
	
	public List<Test> getTestsForFlow(Flow flow) {
		return model.getDeclarations().stream() //
				.filter(d -> d instanceof Test) //
				.map(test -> ((Test) test)) //
				.filter(test -> test.getFlowReference().getRef().getName().equals(flow.getName())) //
				.collect(Collectors.toList()); //
	}
	
	public List<Test> getTests() {
		return model.getDeclarations().stream() //
				.filter(d -> d instanceof Test) //
				.map(test -> ((Test) test)) //
				.collect(Collectors.toList()); //
	}
	
	public List<Loop> getLoops() {
		return model.getDeclarations().stream() //
			.filter(l -> l instanceof Loop) //
			.map(loop -> ((Loop) loop)) //
			.collect(Collectors.toList()); //
	}
	
	public Loop getLoop(String loopId) {
		Optional<Loop> loopOpt = model.getDeclarations().stream() //
			.filter(l -> l instanceof Loop) //
			.map(loop -> ((Loop) loop)) //
			.filter(loop -> loop.getName().equals(loopId)) //
			.findFirst(); //
		
		return loopOpt.isPresent() ? loopOpt.get() : null;
	}
	
	public void deleteLoop(String name) {
		Optional<Loop> loopOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof Loop) //
				.map(loop -> ((Loop) loop)) //
				.filter(loop -> loop.getName().equals(name)) //
				.findFirst(); //
		
		if(loopOpt.isPresent()) {
			model.getDeclarations().remove(loopOpt.get());
		}
	}
	
	public void addBPMNDiagram(BPMNDiagram diagram) {
		model.getDeclarations().add(diagram);
	}

	public void addVariable(Variable variable) {
		model.getDeclarations().add(variable);
	}

	public void addFlowElement(FlowElement flowElement) {
		model.getDeclarations().add(flowElement);
	}

	public void addFlow(Flow flow) {
		model.getDeclarations().add(flow);
	}

	public void addTest(Test test) {
		model.getDeclarations().add(test);
	}
	
	public void addExternalServiceMock(ExternalServiceMock externalServiceMock) {
		model.getDeclarations().add(externalServiceMock);
	}
	
	public void addDelegateMock(DelegateMock delegateMock) {
		model.getDeclarations().add(delegateMock);
	}
	
	public void addExternalTopic(ExternalTopic externalTopic) {
		model.getDeclarations().add(externalTopic);
	}
	
	public void deleteExternalTopic(String name) {
		Optional<ExternalTopic> topicOpt = model.getDeclarations().stream() //
				.filter(d -> d instanceof ExternalTopic) //
				.map(externalTopic -> ((ExternalTopic) externalTopic)) //
				.filter(externalTopic -> externalTopic.getName().equals(name)) //
				.findFirst(); //
		
		if(topicOpt.isPresent()) {
			model.getDeclarations().remove(topicOpt.get());
		}
	}
	
	public void addLoop(Loop loop) {
		model.getDeclarations().add(loop);
	}
	
	public ExternalTopic getExternalTopic(String externalTopicName) {
		Optional<ExternalTopic> findFirst = model.getDeclarations().stream() //
			.filter(decl -> decl instanceof ExternalTopic) //
			.map(externalTopic -> ((ExternalTopic) externalTopic)) //
			.filter(externalTopic -> externalTopic.getName().equals(externalTopicName)) //
			.findFirst();
		
		return findFirst.isPresent() ? findFirst.get() : null;
	}
	
	public List<ExternalTopic> getExternalTopics() {
		return model.getDeclarations().stream() //
				.filter(e -> e instanceof ExternalTopic) //
				.map(e -> ((ExternalTopic) e)) //
				.collect(Collectors.toList()); //
	}
	
	
	public void serialize() throws IOException {
		String serialized = serializer.serialize(model);
		serialized = serialized.replace("   ", " ");
		
		FileWriter myWriter = new FileWriter(dslFile);
		myWriter.write(serialized);
		myWriter.close();
	}

	public class ReferenceComparator implements Comparator<FlowElement> {

		private List<String> ids;

		public ReferenceComparator(List<String> ids) {
			this.ids = ids;
		}

		@Override
		public int compare(FlowElement left, FlowElement right) {
			return Integer.compare(ids.indexOf(left.getName()), ids.indexOf(right.getName()));
		}

	}
}
